import mutex, thread
import sgpem

## @brief This is an adapter class which acts as a proxy
# for user-implemented policies
#
# At runtime, this class will be initialized with the
# user-implemented policy, and then it will proxy the calls
# the policy member functions on a different thread, so
# to ensure asyncronous control.
#
# Instantiated in the C++ code, it is created like:
# @code
# adapter = ScriptAdapter(UserPolicyClass)
# @endcode
#
# The user shouldn't care about this class at all.
class ScriptAdapter :
    ## @var The policy this ScriptAdapter will use for calls
    _policy = None

    ## @var Synchronized return value you can read from C++
    # when a threaded function returns
    _ret_val = None

    ## Var Testable syncronization object
    _g_mutex = mutex.mutex()

    ## @brief Constructor of ScriptAdapter
    #
    # @param self The caller object
    # @param policy A user-implemented class inheriting from Policy.Policy
    def __init__(self, policy):
        self._policy = policy()
        print 'ScriptAdapter for policy ', policy, ' loaded'


    ## @brief Asynchronously call Policy.configure()
    #
    # @param self The caller object
    def async_configure(self):
        self._g_mutex.lock(ScriptAdapter._wrap_configure, self )
        
    def _wrap_configure(self):
        thread.start_new_thread(ScriptAdapter._wrap_configure_callback, (self,))
        
    def _wrap_configure_callback(self):
        # call configure method
        self._policy.configure()
        self._g_mutex.unlock()


    ## @brief Asynchronously call Policy.sort_queue()
    #
    # The queue is asked directly to the C++ sgpem::Scheduler
    # singleton, via SWIG
    #
    # @param self The caller object
    def async_sort_queue(self):
        self._g_mutex.lock(ScriptAdapter._wrap_sort_queue, self)

    def _wrap_sort_queue(self):
        thread.start_new_thread(ScriptAdapter._wrap_sort_queue_callback, (self,))

    def _wrap_sort_queue_callback(self):
        # here we retrieve and pass the ready queue
        queue = sgpem.Scheduler.get_instance().get_ready_queue()
        self._policy.sort_queue(queue)
        self._g_mutex.unlock()


    ## @brief Asynchronously call Policy.is_preemptive()
    #
    # @param self The caller object
    def async_is_preemptive(self):
        self._g_mutex.lock(ScriptAdapter._wrap_is_preemptive, self)

    def _wrap_is_preemptive(self):
        thread.start_new_thread(ScriptAdapter._wrap_is_preemptive_callback, (self,))

    def _wrap_is_preemptive_callback(self):
        self._ret_val = self._policy.is_preemptive()
        self._g_mutex.unlock()
    
    ## @brief Asynchronously call Policy.get_time_slice()
    #
    # @param self The caller object
    def async_get_time_slice(self):
        self._g_mutex.lock(ScriptAdapter._wrap_get_time_slice, self)

    def _wrap_get_time_slice(self):
        thread.start_new_thread(ScriptAdapter._wrap_get_time_slice_callback, (self,))

    def _wrap_get_time_slice_callback(self):
        self._ret_val = self._policy.get_time_slice()
        self._g_mutex.unlock()
    
    ## @brief Return the global shared variable with the methods' last return value
    def get_return_value(self):
        return self._ret_val

    def mutex_test_lock(self):
        return self._g_mutex.test()
